@@ -3,13 +3,14 @@ Google Apps authentication plugin for Trac
 """
 
 from trac.core import *
-from trac.config import Option
-
-from acct_mgr.api import IPasswordStore
+from trac.config import Option, BoolOption
+from trac.env import ISystemInfoProvider
 from trac.perm import IPermissionGroupProvider
 
-import gdata.apps.service
-import gdata.apps.groups.service
+from acct_mgr.api import IPasswordStore
+
+from gdata.apps.service import AppsService
+from gdata.apps.groups.service import GroupsService
 from gdata.service import BadAuthentication, CaptchaRequired
 from gdata.apps.service import AppsForYourDomainException
 
@@ -17,7 +18,7 @@ from gdata.apps.service import AppsForYourDomainException
 class GoogleAppsPasswordStore(Component):
 	"""TracAccountManager Password Store which authenticates against a Google Apps domain"""
 	
-	implements(IPasswordStore, IPermissionGroupProvider)
+	implements(IPasswordStore, IPermissionGroupProvider, ISystemInfoProvider)
 	
 	# Plugin Options
 	opt_key = 'google_apps'
@@ -25,6 +26,8 @@ class GoogleAppsPasswordStore(Component):
 	gapps_admin_username = Option(opt_key, 'admin_username', '', """Username or email with Google Apps admin access""")
 	gapps_admin_secret = Option(opt_key, 'admin_secret', '', """Password for Google Apps admin account""")
 	gapps_group_access = Option(opt_key, 'group_access', '', """Optional Google Apps Group which is exclusively granted access to this Trac site""")
+	#enable_multiple_stores = BoolOption(opt_key, 'enable_multiple_stores', False, """Optional flag to enable use of this plugin alongside other PasswordStore implementations (will slightly increase network overhead)""")
+	
 	
 	def __init__(self):
 		self.env.log.debug('GoogleAppsPasswordStore.__init__()')
@@ -50,19 +53,19 @@ class GoogleAppsPasswordStore(Component):
 	def _get_apps_service(self):
 		"""Build instance of gdata AppsService ready for ProgrammaticLogin()"""
 		email = self._get_admin_email()
-		return gdata.apps.service.AppsService(email=email, domain=self.gapps_domain, password=self.gapps_admin_secret)
+		return AppsService(email=email, domain=self.gapps_domain, password=self.gapps_admin_secret)
 	
 	def _get_groups_service(self):
 		"""Build instance of gdata GroupsService ready for ProgrammaticLogin()"""
 		email = self._get_admin_email()
-		return gdata.apps.groups.service.GroupsService(email=email, domain=self.gapps_domain, password=self.gapps_admin_secret)
+		return GroupsService(email=email, domain=self.gapps_domain, password=self.gapps_admin_secret)
 	
 	def _get_all_users(self):
 		service = self._get_apps_service()
 		service.ProgrammaticLogin()
 		gapps_users_feed = service.RetrieveAllUsers()
 		self.log.debug(gapps_users_feed)
-		users = map(lambda x: x.login.user_name, gapps_users_feed.entry)
+		users = [user.login.user_name for user in gapps_users_feed.entry]
 		
 		# EXPERIMENTAL ...
 		#db = self.env.get_db_cnx()
@@ -77,13 +80,17 @@ class GoogleAppsPasswordStore(Component):
 	def _get_all_users_in_group(self, group, suspended_users=True):
 		service = self._get_groups_service()
 		service.ProgrammaticLogin()
-		gapps_users_feed = service.RetrieveAllMembers(group, suspended_users)
+		gapps_users_feed = service.RetrieveAllMembers(group, suspended_users)  # BUG catch AppsForYourDomainException if group doesn't exist
 		self.log.debug(gapps_users_feed)  # DEBUG
-		emails = map(lambda x: x['memberId'], gapps_users_feed)  # yuck, Groups service gives us email addresses instead of usernames
-		emails = filter(lambda x: x.endswith(self.gapps_domain), emails)  # toss out members outside our domain
-		users = map(lambda x: x.split('@')[0], emails)
+		emails = [user_dict['memberId'] for user_dict in gapps_users_feed]  # Groups service gives us email addresses instead of usernames
+		users = [email.split('@')[0] for email in emails if email.endswith(self.gapps_domain)]  # toss out members outside our domain
 		return users
 	
+	def _user_in_group(self, username, groupname):
+		groups_service = self._get_groups_service()
+		groups_service.ProgrammaticLogin()
+		return groups_service.IsMember(username, groupname)
+
 	
 	# IPasswordStore API
 	
@@ -141,7 +148,7 @@ class GoogleAppsPasswordStore(Component):
 			return None
 		
 		# This should allow the use of multiple PasswordStore implementations at the expense of making the login process slower...
-		#if user not in self.get_users():
+		#if self.enable_multiple_stores and user not in self.get_users():
 		#	self.log.debug('User "%s" is not one of our users, Google Apps plugin falling through.' % (user))
 		#	return None
 		
@@ -149,6 +156,11 @@ class GoogleAppsPasswordStore(Component):
 		service = self._get_apps_service()
 		try:
 			service.ClientLogin(email, password, account_type='HOSTED', source='createtank-tracgoogleappsauthplugin-0.2')
+			
+			if self.gapps_group_access and not self._user_in_group(user, self.gapps_group_access):
+				self.log.debug('User "%s" not in required Google Apps group "%s", login rejected.' % (user, self.gapps_group_access))
+				return False
+			
 			if self.group_cache.has_key(user):
 				self.log.debug('Flushing group cache for user "%s".' % (user))
 				del self.group_cache[user]
@@ -204,7 +216,7 @@ class GoogleAppsPasswordStore(Component):
 			try:
 				service.ProgrammaticLogin()
 				groups_feed = service.RetrieveGroups(username, False)
-				groups = map(lambda x: x['groupName'], groups_feed)
+				groups = [group_dict['groupName'] for group_dict in groups_feed]
 		
 				self.group_cache[username] = groups
 			except AppsForYourDomainException, e:
@@ -213,4 +225,15 @@ class GoogleAppsPasswordStore(Component):
 			
 		self.log.debug(groups)
 		return groups
-		
\ No newline at end of file
+	
+	
+	# ISystemInfoProvider API
+	
+	def get_system_info(self):
+		"""Yield a sequence of `(name, version)` tuples describing the name and
+		version information of external packages used by a component.
+		"""
+		import gdata
+		from trac.util import get_pkginfo
+		yield 'Google Data', get_pkginfo(gdata)['version']
+
